package io.tswf.groovy.bettergroovy.transforms.annotationgstrings

import groovyjarjarasm.asm.Opcodes
import org.codehaus.groovy.ast.ASTNode
import org.codehaus.groovy.ast.AnnotationNode
import org.codehaus.groovy.ast.ClassHelper
import org.codehaus.groovy.ast.ClassNode
import org.codehaus.groovy.ast.FieldNode
import org.codehaus.groovy.ast.expr.ConstantExpression
import org.codehaus.groovy.ast.expr.GStringExpression
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.AbstractASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass
import org.codehaus.groovy.vmplugin.VMPluginFactory

import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy

@GroovyASTTransformationClass(classes = GStringInAnnotationsPreprocessor)
@Retention(RetentionPolicy.SOURCE)
@interface PreprocessGStringsForAnnotations {

}

@GroovyASTTransformation(phase = CompilePhase.CONVERSION)
class GStringInAnnotationsPreprocessor extends AbstractASTTransformation{
    public static final String DELAYED_GSTIRNG_META_KEY = "${GStringInAnnotationsPreprocessor.name}-cached-g-string"
    public static final String IS_TMP_FIELD_META_KEY = "${GStringInAnnotationsPreprocessor.name}-is-tmp"
    @Override
    void visit(ASTNode[] nodes, SourceUnit source) {
        init(nodes, source)
        def cn = nodes.find {it instanceof ClassNode} as ClassNode
        cn.annotations.forEach {delayGStringsInAnnotation(cn, it)}

        cn.allDeclaredMethods.forEach {mn ->
            mn.annotations.forEach {delayGStringsInAnnotation(cn, it)}

            mn.parameters.each {pn ->
                pn.annotations.forEach {delayGStringsInAnnotation(cn, it)}
            }
        }

        cn.declaredConstructors.forEach {mn ->
            mn.annotations.forEach {delayGStringsInAnnotation(cn, it)}

            mn.parameters.each {pn ->
                pn.annotations.forEach {delayGStringsInAnnotation(cn, it)}
            }
        }

        cn.fields.forEach {mn ->
            mn.annotations.forEach {delayGStringsInAnnotation(cn, it)}
        }
    }

    void delayGStringsInAnnotation(ClassNode cn, AnnotationNode annotation) {
        def members = new HashMap<>(annotation.members)
        for (def memberEntry : members) {
            def memberValueExpr = memberEntry.value
            if (memberValueExpr instanceof GStringExpression) {
                def memberGStringValue = memberValueExpr as GStringExpression

                def tmpField = new FieldNode("temp_${UUID.randomUUID().toString().replace("-", "")}", Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC | Opcodes.ACC_TRANSIENT, ClassHelper.make(GString), cn, memberGStringValue)
                cn.addField(tmpField)
                tmpField.putNodeMetaData(IS_TMP_FIELD_META_KEY, true)

                println "Cached GString $memberGStringValue for $annotation.classNode.name in $cn.name!"
                def constant = new ConstantExpression("GString Will Be Here!")
                constant.putNodeMetaData(DELAYED_GSTIRNG_META_KEY, memberGStringValue)
                annotation.members[memberEntry.key] = constant
            }
        }

        VMPluginFactory.getPlugin().configureAnnotation(annotation)
    }
}
